package de.adesso.his.pgbug;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.sql.BatchUpdateException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Properties;

import junit.framework.TestCase;

import org.apache.log4j.Logger;

public class TestDataLoaderPlain extends TestCase {
    private static Logger logger = Logger.getRootLogger();

    private Connection con;

    public static final SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-ss");

    private static final int batchPortionSize = 2;

    public static final int[] colTypes = new int[] { Types.INTEGER, Types.INTEGER,
                                                    Types.FLOAT, Types.SMALLINT, Types.SMALLINT, Types.SMALLINT,
                                                    Types.CHAR, Types.CHAR, Types.CHAR, Types.SMALLINT, Types.CHAR,
                                                    Types.CHAR, Types.CHAR, Types.INTEGER, Types.SMALLINT,
                                                    Types.DATE, Types.FLOAT, Types.SMALLINT, Types.FLOAT,
                                                    Types.CHAR, Types.SMALLINT, Types.CHAR, Types.INTEGER,
                                                    Types.CHAR, Types.CHAR, Types.INTEGER, Types.CHAR,
                                                    Types.SMALLINT, Types.CHAR, Types.CHAR, Types.FLOAT,
                                                    Types.INTEGER, Types.CHAR, Types.INTEGER, Types.CHAR,
                                                    Types.SMALLINT, Types.SMALLINT, Types.SMALLINT, Types.SMALLINT,
                                                    Types.INTEGER, Types.SMALLINT, Types.INTEGER, Types.SMALLINT,
                                                    Types.INTEGER, Types.CHAR, Types.SMALLINT, Types.SMALLINT,
                                                    Types.SMALLINT, Types.CHAR, Types.CHAR, Types.CHAR,
                                                    Types.FLOAT, Types.DATE, Types.CHAR, Types.CHAR, Types.CHAR,
                                                    Types.CHAR, Types.CHAR, Types.CHAR, Types.INTEGER,
                                                    Types.SMALLINT, Types.SMALLINT, Types.CHAR, Types.SMALLINT,
                                                    Types.CHAR, Types.CHAR, Types.CHAR, Types.CHAR, Types.SMALLINT,
                                                    Types.SMALLINT, Types.INTEGER };

    @Override
    public void setUp() throws Exception {
        super.setUp();
        setupConnection();
        createTable();
    }

    private void setupConnection() throws SQLException {
        String url = "jdbc:postgresql://localhost/leerTest";
        Properties props = new Properties();
        props.setProperty("user", "postgres");
        props.setProperty("password", "root");
        con = DriverManager.getConnection(url, props);
    }

    private void createTable() throws SQLException {
        Statement stmt = null;
        try {
            stmt = con.createStatement();
            stmt.executeUpdate(createString);
        } catch (Exception e) {
            logger.error(e);
        } finally {
            if (stmt != null) {
                stmt.close();
            }
        }
    }

    @Override
    protected void tearDown() throws Exception {
        Statement stmt = null;
        try {
            stmt = con.createStatement();
            stmt.executeUpdate("drop table sos_lab_pord");
        } finally {
            if (stmt != null) {
                stmt.close();
            }
        }
    }

	public void testLoad() throws Exception {
		URL file = this.getClass().getResource("/sos_lab_pord_qamuster");
		logger.debug("File: " + file);
		loadTableDataFromFile(new File(file.toURI()), "//", "\\*");
	}

    public void loadTableDataFromFile(File unlFile, String fileCommentMark,
                                      String fileSepStr) throws SQLException, ParseException {

        // Bookkeeping Array: 0 success, 1 failed
        int[] bookkeeping = new int[2];
        bookkeeping[0] = 0;
        bookkeeping[1] = 0;
        FileReader unlReader = null;
        try {
            unlReader = new FileReader(unlFile);
        } catch (FileNotFoundException fe) {
            logger.error("Unl-Datei " + unlFile + " nicht gefunden: " + fe);
        }

        PreparedStatement preparedStatement = null;
        String sql = "INSERT INTO \"sos_lab_pord\" (matrikel_nr, labnr, alter, geschlecht, ca12_staat, hzbart, ch30_fach, ch39_vertief, ch35_ang_abschluss, pversion, schwerpunkt, kz_fach, stort, tid_stg, sem_der_pruefung, d_abg_pruefung, note, abschnitt, fach_sem_zahl, pstatus, prueck, pvermerk, pnr, zeitabschnitt, pform, pordnr, part, schriftlich, ppflicht, modulart, bonus_pord, summe, pktxt, pversuch, pseudonym, studiengang_nr, fach_nr, geburtsjahr, hmkfzkz, hmkfz, semkfzkz, semkfz, hzbkfzkz, hzbkfz, hrst, kz_rueck_beur_ein, klinsem, hssem, hzbart_int, ch27_grund_beurl, ch62_grund_exmatr, hzbnote, d_exmatr, stufrm, ch35_ang_abschluss_stg, ch30_fach_stg, ch39_vertief_stg, schwerpunkt_stg, kz_fach_stg, pversion_stg, sem_einschreibung, sem_1fs, pdum, sem_wechsel, ch35_ang_abschluss_einschr, ch30_fach_einschr, ch35_ang_abschluss_1fs, ch30_fach_1fs, fachsem_beurl_einschr, fachsem_beurl_1fs, sourcesystem) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
        preparedStatement = con.prepareStatement(sql);

        if (unlReader != null) {
            BufferedReader in = new BufferedReader(unlReader);
            int lineCount = 1;
            while (true) {
                String line = null;
                line = getNextLine(in, lineCount);
                if (line == null) {
                    break;
                }

                if ((bookkeeping[0] + bookkeeping[1]) % 100 == 0) {
                    String msg = bookkeeping[0] + " Zeilen geladen";
                    if (bookkeeping[1] > 0) {
                        msg += "; <strong><font color=\"red\">"
                               + bookkeeping[1]
                               + " fehlgeschlagen</font></strong>";
                    }
                    msg += "; <strong>IN ARBEIT ...</strong>";
                }


                if ((line.trim().length() > 0)
                    && !line.startsWith(fileCommentMark)) {
                    String[] colValues = line.split(fileSepStr);
                    if (preparedStatement != null) {
                        putToStatement(preparedStatement, colValues);
                        try {
                            preparedStatement.addBatch();
                        } catch (SQLException e) {
                            logger.error(e, e);
                        }
                    }
                } // endif 
                lineCount++;
                if (lineCount % batchPortionSize == 0) {
                    // logger.info("------------- lineCount=" + lineCount);
                    executeBatchInsert(bookkeeping, preparedStatement, sql,
                                    lineCount);
                }
            } // endwhile 

            // execute the batch
            executeBatchInsert(bookkeeping, preparedStatement, sql, lineCount);

            try {
                in.close();
            } catch (IOException ioe) {
                logger.error("Fehler beim Schließen der Unl-Datei: " + ioe);
            }
        } 

        String msg = bookkeeping[0] + " Zeilen Default-Daten geladen aus '"
                     + unlFile + "'";
        if (bookkeeping[1] > 0) {
            msg += "; <strong><font color=\"red\">" + bookkeeping[1]
                   + " fehlgeschlagen</font></strong>";
        }
        logger.info(msg);
    }


    /**
     * @param in
     * @param lineCount
     * @return line
     */
    public static String getNextLine(BufferedReader in, int lineCount) {
        String line = null;
        try {
            line = in.readLine();
            while (line != null && line.endsWith("\\") && in.ready()) {
                String nextpart = in.readLine();
                if (nextpart == null) {
                    break;
                }
                line = line + "\r\n" + nextpart;
            }
        } catch (IOException ioe) {
            logger.error("Fehler beim Lesen von Zeile " + lineCount + ": "
                         + ioe);
        }
        return line;
    }
    

    /**
     * @param preparedStatement
     * @param colValues
     * @throws SQLException
     * @throws ParseException
     */
    public static void putToStatement(PreparedStatement preparedStatement, String[] colValues) throws SQLException, ParseException {
        for (int i = 0; i < colValues.length; i++) {
            String val = colValues[i];

            if (val.isEmpty()) {
                if (i >= colTypes.length)
                {
                    continue;
                }
                preparedStatement.setNull(i + 1, colTypes[i]);
                continue;
            }

            switch (colTypes[i]) {
            case Types.INTEGER:
            case Types.SMALLINT:
            case Types.FLOAT:
                try {
                    //TODO: Not Working
                    preparedStatement.setObject(i + 1, new Double(val));

                    //Working
                    //preparedStatement.setObject(i + 1, new Double(val), colTypes[i]);
                } catch (NumberFormatException e) {
                    e.printStackTrace();
                }
                break;
            case Types.DATE:
                preparedStatement.setObject(i + 1, new java.sql.Date(format.parse(val).getTime()));
                break;

            default:
                preparedStatement.setObject(i + 1, val);
                break;
            }
        }
    }

    /**
     * @param bookkeeping
     * @param preparedStatement
     * @param sql
     * @param lineCount
     */
    private static void executeBatchInsert(int[] bookkeeping,
                                           PreparedStatement preparedStatement, String sql, int lineCount) {
        if (preparedStatement != null) {
            try {
                int[] res = preparedStatement.executeBatch();
                preparedStatement.clearBatch();
                bookkeeping[0] = bookkeeping[0] + res.length;
            } catch (BatchUpdateException e) {
                int[] res = e.getUpdateCounts();
                for (int lineRes : res) {
                    if (lineRes == Statement.EXECUTE_FAILED) {
                        bookkeeping[1]++;
                    } else {
                        bookkeeping[0]++;
                    }
                }
                logger.error("copyRows():12: Fehler beim Laden im Batch " + sql
                             + ": ");
                SQLException sqlException = e;
                do {
                    logger.error(sqlException.toString(), sqlException);
                    sqlException = sqlException.getNextException();
                } while (sqlException != null);
            } catch (SQLException e) {
                logger.error("copyRows():12: Fehler beim Laden im Batch " + sql
                             + ": ");
                SQLException sqlException = e;
                do {
                    logger.error(sqlException.toString(), sqlException);
                    sqlException = sqlException.getNextException();
                } while (sqlException != null);
                bookkeeping[0] = 0;
                bookkeeping[1] = lineCount;
            }
        }
    }

    private static String createString = "CREATE TABLE sos_lab_pord\r\n"
                                         + "(\r\n" + "  matrikel_nr integer,\r\n" + "  labnr integer,\r\n"
                                         + "  alter numeric(14,2),\r\n" + "  geschlecht smallint,\r\n"
                                         + "  ca12_staat smallint,\r\n" + "  hzbart smallint,\r\n"
                                         + "  ch30_fach character(10),\r\n"
                                         + "  ch39_vertief character(10),\r\n"
                                         + "  ch35_ang_abschluss character(10),\r\n"
                                         + "  pversion smallint,\r\n" + "  schwerpunkt character(10),\r\n"
                                         + "  kz_fach character(10),\r\n" + "  stort character(10),\r\n"
                                         + "  tid_stg integer,\r\n"
                                         + "  sem_der_pruefung smallint NOT NULL,\r\n"
                                         + "  d_abg_pruefung date,\r\n" + "  note numeric(3,2),\r\n"
                                         + "  abschnitt smallint,\r\n"
                                         + "  fach_sem_zahl numeric(10,2),\r\n"
                                         + "  pstatus character(10),\r\n" + "  prueck smallint,\r\n"
                                         + "  pvermerk character(10),\r\n" + "  pnr integer,\r\n"
                                         + "  zeitabschnitt character(10),\r\n"
                                         + "  pform character(10),\r\n" + "  pordnr integer,\r\n"
                                         + "  part character(10),\r\n" + "  schriftlich smallint,\r\n"
                                         + "  ppflicht character(2),\r\n" + "  modulart character(10),\r\n"
                                         + "  bonus_pord double precision,\r\n" + "  summe integer,\r\n"
                                         + "  pktxt character(255),\r\n" + "  pversuch integer,\r\n"
                                         + "  pseudonym character varying(255),\r\n"
                                         + "  studiengang_nr smallint,\r\n" + "  fach_nr smallint,\r\n"
                                         + "  geburtsjahr smallint,\r\n" + "  hmkfzkz smallint,\r\n"
                                         + "  hmkfz integer,\r\n" + "  semkfzkz smallint,\r\n"
                                         + "  semkfz integer,\r\n" + "  hzbkfzkz smallint,\r\n"
                                         + "  hzbkfz integer,\r\n" + "  hrst character(10),\r\n"
                                         + "  kz_rueck_beur_ein smallint,\r\n" + "  klinsem smallint,\r\n"
                                         + "  hssem smallint,\r\n" + "  hzbart_int character(10),\r\n"
                                         + "  ch27_grund_beurl character(10),\r\n"
                                         + "  ch62_grund_exmatr character(10),\r\n"
                                         + "  hzbnote numeric(6,2),\r\n" + "  d_exmatr date,\r\n"
                                         + "  stufrm character(10),\r\n"
                                         + "  ch35_ang_abschluss_stg character(10),\r\n"
                                         + "  ch30_fach_stg character(10),\r\n"
                                         + "  ch39_vertief_stg character(10),\r\n"
                                         + "  schwerpunkt_stg character(10),\r\n"
                                         + "  kz_fach_stg character(10),\r\n"
                                         + "  pversion_stg integer,\r\n"
                                         + "  sem_einschreibung smallint,\r\n" + "  sem_1fs smallint,\r\n"
                                         + "  pdum character(10),\r\n" + "  sem_wechsel smallint,\r\n"
                                         + "  ch35_ang_abschluss_einschr character(10),\r\n"
                                         + "  ch30_fach_einschr character(10),\r\n"
                                         + "  ch35_ang_abschluss_1fs character(10),\r\n"
                                         + "  ch30_fach_1fs character(10),\r\n"
                                         + "  fachsem_beurl_einschr smallint,\r\n"
                                         + "  fachsem_beurl_1fs smallint,\r\n"
                                         + "  sourcesystem integer\r\n" + ")\r\n" + "WITH (\r\n"
                                         + "  OIDS=FALSE\r\n" + ")";

}
